#include <config.h>

#include <stdio.h>
#include <glib.h>
#include <glib-object.h>
#include <assert.h>

#include "btctl.h"
#include "btctlimpl.h"
#include "btctl-marshal.h"

static gpointer         parent_class = NULL;

/* gtk object prototypes */

static void btctl_class_init (BtctlClass *class);
static void btctl_init (Btctl *bc);
static void btctl_finalize(GObject *obj);

/* gtk signal defns */

enum {
	ADD_DEVICE_SIGNAL,
	DEVICE_NAME_SIGNAL,
	STATUS_CHANGE_SIGNAL,
	ADD_DEVICE_SERVICE_SIGNAL,
	LAST_SIGNAL
};

static gint btctl_signals[LAST_SIGNAL] = { 0 } ;

/* Btctl functions */

guint btctl_get_type()
{
	static guint model_type = 0;
	if (!model_type) {
		GTypeInfo model_info =
		{
			sizeof(BtctlClass),
	    NULL, /* base init */
	    NULL, /* base finalize */
			(GClassInitFunc) btctl_class_init,
	    NULL, /* class finalize */
	    NULL, /* class data */
	    sizeof(Btctl),
	    1, /* n_preallocs */
			(GInstanceInitFunc) btctl_init,
		};
		model_type = g_type_register_static(G_TYPE_OBJECT,
		"btctl",  &model_info, 0);
	}
	return model_type;
}


static void btctl_class_init (BtctlClass *klass) {
	GObjectClass *object_class;

	parent_class = g_type_class_ref(G_TYPE_OBJECT);

	object_class=(GObjectClass*)klass;

	btctl_signals[ADD_DEVICE_SIGNAL] = g_signal_new (
	    "add_device",
	    G_OBJECT_CLASS_TYPE(klass),
			G_SIGNAL_RUN_FIRST,
			G_STRUCT_OFFSET(BtctlClass, add_device),
	    NULL /* accu */, NULL,
			btctl_marshal_VOID__STRING_UINT,
			G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_UINT);

    btctl_signals[DEVICE_NAME_SIGNAL] = g_signal_new (
	    "device_name",
	    G_OBJECT_CLASS_TYPE(klass),
			G_SIGNAL_RUN_FIRST,
			G_STRUCT_OFFSET(BtctlClass, device_name),
	    NULL /* accu */, NULL,
			btctl_marshal_VOID__STRING_STRING,
			G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);


	btctl_signals[STATUS_CHANGE_SIGNAL] = g_signal_new (
	    "status_change",
	    G_OBJECT_CLASS_TYPE(klass),
			G_SIGNAL_RUN_FIRST,
			G_STRUCT_OFFSET(BtctlClass, status_change),
	    NULL /* accu */, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1, G_TYPE_INT);

	btctl_signals[ADD_DEVICE_SERVICE_SIGNAL] =
	  g_signal_new (
			"add_device_service",
			G_OBJECT_CLASS_TYPE(klass),
			G_SIGNAL_RUN_FIRST,
			G_STRUCT_OFFSET(BtctlClass, add_device_service),
			NULL, NULL,
			btctl_marshal_VOID__STRING_STRING_UINT_UINT,
			G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_STRING,
			G_TYPE_UINT, G_TYPE_UINT);

	object_class->finalize = btctl_finalize;

	klass->add_device=NULL;
}

static void btctl_init (Btctl *bc) {
	bc->ctl=btctl_impl_init();
    /* TODO: report error if bc->ctl < 0 */
}

GObject* btctl_new() {
	return g_object_new (btctl_get_type(), NULL);
}

/* local private methods */


static void btctl_finalize(GObject *obj) {
	Btctl *bc;

	bc=BTCTL(obj);

	/* custom finalize stuff goes here
	 */
    btctl_impl_free(bc);
	/* call superclass' destructor */
    G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void btctl_emit_add_device(Btctl *bc, gchar  *bdaddr, guint clsid) {
    g_signal_emit (G_OBJECT (bc),
		btctl_signals[ADD_DEVICE_SIGNAL],
		0,
		bdaddr, clsid);
}

static void btctl_emit_device_name(Btctl *bc, gchar  *bdaddr, gchar *name) {
    g_signal_emit (G_OBJECT (bc),
		btctl_signals[DEVICE_NAME_SIGNAL],
		0,
		bdaddr, name);
}


static void btctl_emit_status_change(Btctl *bc, gint status) {
    g_signal_emit (G_OBJECT (bc),
		btctl_signals[STATUS_CHANGE_SIGNAL],
		0,
		status);
}

static void btctl_emit_device_service(Btctl *bc, gchar *bdaddr,
				      gchar *name, guint classid, guint port) {
  g_signal_emit (G_OBJECT(bc),
		 btctl_signals[ADD_DEVICE_SERVICE_SIGNAL],
		 0,
		 bdaddr, name, classid, port);
}

/* exported but private */

void btctl_report_status(Btctl *bc, gint status) {
    btctl_emit_status_change(bc, status);
}

void btctl_got_device(Btctl *bc, gchar *bdaddr, guint clsid) {
    btctl_emit_add_device(bc, bdaddr, clsid);
}

void btctl_got_device_name(Btctl *bc, gchar *bdaddr, gchar *name) {
    btctl_emit_device_name(bc, bdaddr, name);
}

void btctl_got_device_service(Btctl *bc, gchar *bdaddr,
			      gchar *name, guint classid, guint port) {
  btctl_emit_device_service(bc, bdaddr, name, classid, port);
}

/* substantial public methods */

void btctl_discover_devices(Btctl *bc)
{
    btctl_impl_cmd_scan(bc);
}

void btctl_list_rfcomm_connections(Btctl *bc)
{
  btctl_impl_list_rfcomm_connections(bc);
}

int btctl_get_established_rfcomm_connection(Btctl *bc, gchar *bdstr,
											guint channel)
{
  return btctl_impl_get_established_rfcomm_connection(bc, bdstr, channel);
}

int btctl_establish_rfcomm_connection(Btctl *bc, gchar *bdstr, guint channel)
{
  return btctl_impl_establish_rfcomm_connection(bc, bdstr, channel);
}
