
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>



#include <termios.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <asm/types.h>
#include <netinet/in.h>


#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/rfcomm.h>
#undef RFCOMM_PSM
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

#include <glib.h>


#include "btctl.h"
#include "btctlimpl.h"


struct search_context {
    char        *svc;       /* Service */
    uuid_t      group;      /* Browse group */
    int     tree;       /* Display full attribute tree */
    uint32_t    handle;     /* Service record handle */
};

struct report_context {
  Btctl *bc;
  char *id;
  guint chan;
};

typedef int (*handler_t) (Btctl *bc, bdaddr_t *bdaddr, struct search_context *arg);

static int do_browse(Btctl *bc, bdaddr_t *bdaddr, void *arg);

static int do_search(Btctl *bc, bdaddr_t *bdaddr, struct search_context *context);

void fire_got_device(char *devaddr);
void fire_got_device_name(char *devaddr, char *name);

void fire_got_device(char *devaddr) {
    printf("DEVICE: %s\n", devaddr);
}

void fire_got_device_name(char *devaddr, char *name) {
    printf("NAME  : %s - %s\n", devaddr, name);
}


static void 
print_service_class(void *value, void *userData)
{
	char ServiceClassUUID_str[MAX_LEN_SERVICECLASS_UUID_STR];
	char UUID_str[MAX_LEN_UUID_STR];

	uuid_t *uuid = (uuid_t *)value;

	sdp_uuid2strn(uuid, UUID_str, MAX_LEN_UUID_STR);
	sdp_svclass_uuid2strn(uuid, ServiceClassUUID_str, MAX_LEN_SERVICECLASS_UUID_STR);
	printf("  \"%s\" (0x%s)\n", ServiceClassUUID_str, UUID_str);
}


static void 
btctl_impl_report_class(void *value, void *userData)
{
  struct report_context *con=(struct report_context *)userData;
  Btctl *bc=con->bc;

  char ServiceClassUUID_str[MAX_LEN_SERVICECLASS_UUID_STR];
  char UUID_str[MAX_LEN_UUID_STR];
  
  uuid_t *uuid = (uuid_t *)value;
  
  sdp_uuid2strn(uuid, UUID_str, MAX_LEN_UUID_STR);
  sdp_svclass_uuid2strn(uuid, ServiceClassUUID_str, MAX_LEN_SERVICECLASS_UUID_STR);
  
  btctl_got_device_service(bc, con->id, ServiceClassUUID_str,
						   (guint)(uuid->value.uuid16),
						   con->chan);
}


static void 
print_service_desc(void *value, void *user)
{
	char str[MAX_LEN_PROTOCOL_UUID_STR];
	char UUID_str[MAX_LEN_UUID_STR];
	guint *chan=(guint*)user;

	sdp_data_t *p = (sdp_data_t *)value, *s;
	int i = 0, proto = 0;
	
	for (; p; p = p->next, i++) {
		switch (p->dtd) {
		case SDP_UUID16:
		case SDP_UUID32:
		case SDP_UUID128:
			sdp_uuid2strn(&p->val.uuid, UUID_str, MAX_LEN_UUID_STR);
			sdp_proto_uuid2strn(&p->val.uuid, str, sizeof(str));
			proto = sdp_uuid_to_proto(&p->val.uuid);
			printf("  \"%s\" (0x%s)\n", str, UUID_str);
			break;
		case SDP_UINT8:
		  if (proto == RFCOMM_UUID) {
			printf("    Channel: %d\n", p->val.uint8);
			*chan=(guint)p->val.uint8;
		  } else {
			printf("    uint8: 0x%x\n", p->val.uint8);
		  }
		  break;
		case SDP_UINT16:
			if (proto == L2CAP_UUID) {
				if (i == 1)
					printf("    PSM: %d\n", p->val.uint16);
				else
					printf("    Version: 0x%04x\n", p->val.uint16);
			} else if (proto == BNEP_UUID)
				if (i == 1)
					printf("    Version: 0x%04x\n", p->val.uint16);
				else
					printf("    uint16: 0x%x\n", p->val.uint16);
			else
				printf("    uint16: 0x%x\n", p->val.uint16);
			break;
		case SDP_SEQ16:
			printf("    SEQ16:");
			for (s = p->val.dataseq; s; s = s->next)
				printf(" %x", s->val.uint16);
			printf("\n");
			break;
		default:
			printf("    FIXME: dtd=0%x\n", p->dtd);
			break;
		}
	}
}

/*
static void 
print_lang_attr(void *value, void *user)
{
	sdp_lang_attr_t *lang = (sdp_lang_attr_t *)value;
	printf("  code_ISO639: 0x%02x\n", lang->code_ISO639);
	printf("  encoding:    0x%02x\n", lang->encoding);
	printf("  base_offset: 0x%02x\n", lang->base_offset);
}
*/

static void 
print_access_protos(void *value, void *userData)
{
	sdp_list_t *protDescSeq = (sdp_list_t *)value;
	sdp_list_foreach(protDescSeq, print_service_desc, userData);
}

/*
static void 
print_profile_desc(void *value, void *userData)
{
	sdp_profile_desc_t *desc = (sdp_profile_desc_t *)value;
	char str[MAX_LEN_PROFILEDESCRIPTOR_UUID_STR];
	char UUID_str[MAX_LEN_UUID_STR];

	sdp_uuid2strn(&desc->uuid, UUID_str, MAX_LEN_UUID_STR);
	sdp_profile_uuid2strn(&desc->uuid, str, MAX_LEN_PROFILEDESCRIPTOR_UUID_STR);

	printf("  \"%s\" (0x%s)\n", str, UUID_str);
	if (desc->version)
		printf("    Version: 0x%04x\n", desc->version);
}*/

/*
 * Parse a SDP record in user friendly form.
 */
static void 
print_service_attr(Btctl *bc, char *id, sdp_record_t *rec)
{
	sdp_list_t *list = 0, *proto = 0;
	struct report_context con;

	sdp_record_print(rec);
	memset((void*)&con, 0, sizeof(struct report_context));
	con.id=id;
	con.bc=bc;

	printf("Service RecHandle: 0x%x\n", rec->handle);
	if (sdp_get_access_protos(rec, &proto) == 0) {
		printf("Protocol Descriptor List:\n");
		sdp_list_foreach(proto, print_access_protos, &con.chan);
		sdp_list_free(proto, (sdp_free_func_t)sdp_data_free);
	}
	if (sdp_get_service_classes(rec, &list) == 0) {
		printf("Service Class ID List:\n");
		sdp_list_foreach(list, print_service_class, 0);
		sdp_list_foreach(list, btctl_impl_report_class, &con);
		sdp_list_free(list, free);
	}
	/*
	if (sdp_get_lang_attr(rec, &list) == 0) {
		printf("Language Base Attr List:\n");
		sdp_list_foreach(list, print_lang_attr, 0);
		sdp_list_free(list, free);
	}
	if (sdp_get_profile_descs(rec, &list) == 0) {
	printf("Profile Descriptor List:\n");
	sdp_list_foreach(list, print_profile_desc, 0);
	sdp_list_foreach(list, btctl_impl_report_profile, &con);
	sdp_list_free(list, free);
	}
*/
}

// TODO:
//   the SDP library probably isn't threadsafe, so we should
//   implement a static mutex for when this method is called.

/*

		btctl_got_device_service(bc, str,
								 btctl_get_service_name(uuid->value.uuid16),
								 (guint)uuid->value.uuid16, theport);
*/

void btctl_impl_cmd_scan(Btctl *bc)
{
  inquiry_info *info=NULL;
  int num_rsp, length, flags;
  bdaddr_t bdaddr;
  char name[248];
  int i, dd;
  int dev_id=0;
  
  length  = 8;  /* ~10 seconds */
  num_rsp = 10;
  flags = 0;
  
  dev_id=hci_devid("hci0");

  if (dev_id < 0) {
    dev_id = hci_get_route(&bdaddr);
    if (dev_id < 0) {
      perror("HCI device not available.");
      /* TODO: report this status to the app */
      return;
    }
  }
  
  printf("dev_id is %d\n", dev_id);

  /*  
  if (sdp_init()<0) {
	perror("Can't init SDP.");

	return;
  }
  */

  btctl_report_status(bc, BTCTL_STATUS_SCANNING);
  
  num_rsp = hci_inquiry(dev_id, length, num_rsp, NULL, &info, flags);
  
  if (num_rsp < 0) {
    perror("Inquiry failed.");
    return;
  }

  printf("Got %d responses.\n", num_rsp);

  for (i = 0; i < num_rsp; i++) {
    baswap(&bdaddr, &(info+i)->bdaddr);
    printf("\t%s\tclock offset: 0x%4.4x\tclass: 0x%2.2x%2.2x%2.2x\n",
	   batostr(&bdaddr), (info+i)->clock_offset,
	   (info+i)->dev_class[2],
	   (info+i)->dev_class[1],
	   (info+i)->dev_class[0]);
    btctl_got_device(bc, batostr(&bdaddr), (info+i)->dev_class[0]|
					 ((info+i)->dev_class[1]<<8)|
					 ((info+i)->dev_class[2]<<16));
  }
  
  btctl_report_status(bc, BTCTL_STATUS_GETTING_NAMES);
  dd = hci_open_dev(dev_id);
  if (dd>=0) {
    for (i = 0; i < num_rsp; i++) {
      memset(name, 0, sizeof(name));
      if (hci_remote_name(dd, &(info+i)->bdaddr,
			  sizeof(name), name, 100000) < 0)
		strcpy(name, "n/a");
      
      baswap(&bdaddr, &(info+i)->bdaddr);
      
      btctl_got_device_name(bc, batostr(&bdaddr), name);
    }

    btctl_report_status(bc, BTCTL_STATUS_GETTING_SERVICES);
    close(dd);
    for(i=0; i<num_rsp; i++) {
      // baswap(&bdaddr, &(info+i)->bdaddr);
      
      do_browse(bc, &((info+i)->bdaddr), NULL);
    }
  } else {
    perror("HCI device open failed.");
    /* TODO: report this status somehow */
  }
  btctl_report_status(bc, BTCTL_STATUS_COMPLETE);
  free(info);
  /*  sdp_shutdown(); */
}

/*
 * Perform an inquiry and search/browse all peer found.
 */
static void 
inquiry(Btctl *bc, handler_t handler, void *arg)
{
    inquiry_info ii[20];
    uint8_t count = 0;
    int i;
 
    printf("Inquiring ...\n");
    if (sdp_general_inquiry(ii, 20, 8, &count) < 0) {
        printf("Inquiry failed\n");
        return;
    }
    for (i=0; i<count; i++)
        handler(bc, &ii[i].bdaddr, arg);
}


/*
 * Search for a specific SDP service
 */
static int 
do_search(Btctl *bc, bdaddr_t *bdaddr, struct search_context *context)
{
	sdp_list_t *attrid, *search, *seq, *next;
	uint32_t range = 0x0000ffff;
	char str[20];
	sdp_session_t *sess;

	if (!bdaddr) {
		inquiry(bc, do_search, context);
		return 0;
	}
	sess = sdp_connect(BDADDR_ANY, bdaddr, 0);
	ba2str(bdaddr, str);
	if (!sess) {
		printf("Failed to connect to SDP server on %s\n", str);
		return -1;
	}
	if (context->svc)
		printf("Searching for %s on %s ...\n", context->svc, str);
	else
		printf("Browsing %s ...\n", str);

	attrid = sdp_list_append(0, &range);
	search = sdp_list_append(0, &context->group);
	if (sdp_service_search_attr_req(sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
		printf("Service Search failed\n");
		sdp_close(sess);
		return -1;
	}
	sdp_list_free(attrid, 0);
	sdp_list_free(search, 0);

	for (; seq; seq = next) {
		sdp_record_t *rec = (sdp_record_t *) seq->data;
		struct search_context sub_context;

		if (context->tree) {
			/* Display full tree */
			/* sdp_printf_service_attr(rec); */
		} else {
			/* Display user friendly form */
			/* btlctl_impl_report_service_attr(rec);*/

		  print_service_attr(bc, str, rec);
		}
		printf("\n");
		
		if (sdp_get_group_id(rec, &sub_context.group) != -1) {
			/* Set the subcontext for browsing the sub tree */
			memcpy(&sub_context, context, sizeof(struct search_context));
			/* Browse the next level down if not done */
			if (sub_context.group.value.uuid16 != context->group.value.uuid16)
				do_search(bc, bdaddr, &sub_context);
		}
		next = seq->next;
		free(seq);
		sdp_record_free(rec);
	}
	sdp_close(sess);
	return 0;
}

static int 
do_browse(Btctl *bc, bdaddr_t *bdaddr, void *arg)
{
  struct search_context context;

  memset(&context, '\0', sizeof(struct search_context));
  /* We want to browse the top-level/root */
  sdp_uuid16_create(&(context.group), PUBLIC_BROWSE_GROUP);
  if (bdaddr) 
	return do_search(bc, bdaddr, &context);
  return do_search(bc, 0, &context);
}

int 
btctl_impl_init() 
{
  /* ret value of <0 is an error */
  return 1;
}

void 
btctl_impl_free(Btctl *bc) 
{
 
}

static char *rfcomm_state[] = {
	"unknown",
	"connected",
	"clean",
	"bound",
	"listening",
	"connecting",
	"connecting",
	"config",
	"disconnecting",
	"closed"
};

static char *rfcomm_flagstostr(uint32_t flags)
{
	char str[100];
	str[0] = 0;

	strcat(str, "[");

	if (flags & (1 << RFCOMM_REUSE_DLC))
		strcat(str, "reuse-dlc ");

	if (flags & (1 << RFCOMM_RELEASE_ONHUP))
		strcat(str, "release-on-hup ");

	if (flags & (1 << RFCOMM_TTY_ATTACHED))
		strcat(str, "tty-attached");
	
	strcat(str, "]");
	return g_strdup(str);
}

int
btctl_impl_establish_rfcomm_connection(Btctl *bc, gchar *bdstr, guint channel)
{
  bdaddr_t bdaddr;
  int dev=-1;
  struct rfcomm_dev_req *req;
  struct rfcomm_dev_list_req *dl;
  struct sockaddr_rc laddr, raddr;
  int sk, alen;
  int ctl;

  ctl=socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM);
  if (!ctl) {
 	g_warning(_("Can't open RFCOMM control socket"));
 	return -1;
  }
  
  dl=g_malloc(sizeof(*dl) + RFCOMM_MAX_DEV * sizeof(struct rfcomm_dev_info));
  if (!dl) {
 	g_warning(_("Can't allocate memory"));
 	close(ctl);
 	return -1;
  }
  dl->dev_num=RFCOMM_MAX_DEV;
  
  if (ioctl(ctl, RFCOMMGETDEVLIST, (void *) dl) < 0) {
 	g_warning(_("Can't get device list"));
 	g_free(dl);
 	close(ctl);
 	return -1;
  }
  close(ctl);

  req=g_malloc0(sizeof(*req));
 
  str2ba(bdstr, &bdaddr);
  
  bacpy(&req->dst, &bdaddr);
  
  laddr.rc_family = AF_BLUETOOTH;
  bacpy(&laddr.rc_bdaddr, BDADDR_ANY);
  laddr.rc_channel = 0;
  raddr.rc_family = AF_BLUETOOTH;
  bacpy(&raddr.rc_bdaddr, &bdaddr);
  raddr.rc_channel=channel;

  if ((sk = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
	g_warning("Can't create RFCOMM socket");
	return -1;
  }
  
  if (bind(sk, (struct sockaddr *)&laddr, sizeof(laddr)) < 0) {
	g_warning("Can't bind RFCOMM socket");
	close(sk);
	return -1;
  }
  
  if (connect(sk, (struct sockaddr *)&raddr, sizeof(raddr)) < 0) {
	g_warning("Can't connect RFCOMM socket");
	close(sk);
	return -1;
  }
  
  alen = sizeof(laddr);
  if (getsockname(sk, (struct sockaddr *)&laddr, &alen) < 0) {
	g_warning("Can't get RFCOMM socket name");
	close(sk);
	return -1;
  }
  
  bacpy(&req->src, &laddr.rc_bdaddr);
  bacpy(&req->dst, &raddr.rc_bdaddr);
  req->dev_id=dl->dev_num;
  req->flags=(1 << RFCOMM_REUSE_DLC) | (1 << RFCOMM_RELEASE_ONHUP);;
  req->channel=raddr.rc_channel;
  
  if ((dev = ioctl(sk, RFCOMMCREATEDEV, req)) < 0) {
	g_warning("Can't create RFCOMM TTY");
	close(sk);
	return -1;
  }

  close(sk);
  g_free(dl);
  g_free(req);


  return dev;
}

/*
  returns the port number of an established rfcomm connection to
  a device.
  returns -1 if nothing available, -2 if port available but connected.
*/

int
btctl_impl_get_established_rfcomm_connection(Btctl *bc, gchar *bdstr,
											 guint channel) 
{
  bdaddr_t bdaddr;
  struct rfcomm_dev_list_req *dl;
  struct rfcomm_dev_info *di;
  int i;
  int dev=-1;
  int ctl=socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM);

  str2ba(bdstr, &bdaddr);

  if (!ctl) {
	g_error(_("Can't open RFCOMM control socket"));
	return -1;
  }

  dl=g_malloc(sizeof(*dl) + RFCOMM_MAX_DEV * sizeof(*di));
  if (!dl) {
	g_warning(_("Can't allocate memory"));
	close(ctl);
	return -1;
  }
  dl->dev_num=RFCOMM_MAX_DEV;
  di=dl->dev_info;
  
  if (ioctl(ctl, RFCOMMGETDEVLIST, (void *) dl) < 0) {
	g_warning(_("Can't get device list"));
	g_free(dl);
	close(ctl);
	return -1;
  }
  
  for(i=0; i<dl->dev_num && dev<0; i++, di++) {
	/** TODO: btctl needs to know which interface it's meant to use:
		it can't just use anyone it likes. Until then we'll make the
	    (wrong) assumption we're on a one-device setup and we'll always
	    get the device we need */
	/* if (bacmp(&di->src, BDADDR_ANY) == 0) { */
	  if (bacmp(&di->dst, &bdaddr)==0 && 
		  (channel==0 || channel==di->channel)) {
		if ((di->flags&(1<<RFCOMM_TTY_ATTACHED))==0) {
		  dev=i;
		} else {
		  /* if a candidate connection exists, but it's in use
			 then time to return -2 */
		  dev=-2;
		}
	  }
	  /* } */
  }
  g_free(dl);
  close(ctl);

  return dev;
}

void
btctl_impl_list_rfcomm_connections(Btctl *bc) 
{
  int ctl=socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM);
  struct rfcomm_dev_list_req *dl;
  struct rfcomm_dev_info *di;
  gchar src[18], dst[18], addr[40];

  int i;

  if (!ctl) {
	g_error(_("Can't open RFCOMM control socket"));
	return;
  }

  dl=g_malloc(sizeof(*dl) + RFCOMM_MAX_DEV * sizeof(*di));
  if (!dl) {
	g_error(_("Can't allocate memory"));
	close(ctl);
	return;
  }
  dl->dev_num=RFCOMM_MAX_DEV;
  di=dl->dev_info;
  
  if (ioctl(ctl, RFCOMMGETDEVLIST, (void *) dl) < 0) {
	g_error("Can't get device list");
	g_free(dl);
	close(ctl);
	return;
  }
  
  for(i=0; i<dl->dev_num; i++, di++) {
	gchar *fs=NULL;
	ba2str(&di->src, src); ba2str(&di->dst, dst);

	if (bacmp(&di->src, BDADDR_ANY) == 0)
		sprintf(addr, "%s", dst);
	else
		sprintf(addr, "%s -> %s", src, dst);

	if (di->flags) 
	  fs=rfcomm_flagstostr(di->flags);
	
	g_message("rfcomm%d: %s channel %d %s %s\n",
			  di->id, addr, di->channel,
			  rfcomm_state[di->state], 
			  di->flags ? fs : "");

	if (fs)
	  g_free(fs);
  }

  g_free(dl);
  close(ctl);
}
